# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import unittest
import zipfile
import tarfile
import os
import shutil
import tempfile

from ..unpack import top_level


class TopLevelTests(unittest.TestCase):

    def setUp(self):
        self._tmp_path = os.path.join(tempfile.gettempdir(), 'pigar/')
        os.mkdir(self._tmp_path)

    def tearDown(self):
        shutil.rmtree(self._tmp_path)

    def test_zip(self):
        # .whl and .egg both are .zip file.
        zip_path = os.path.join(self._tmp_path, 'pigar-1.1.zip')
        zf = zipfile.ZipFile(
            zip_path, mode='w', compression=zipfile.ZIP_DEFLATED)
        try:
            zf.writestr('pigar-info/top_level.txt', 'pigar\npigar/tests')
        finally:
            zf.close()
        with open(zip_path, 'rb') as f:
            self.assertListEqual(
                top_level(zip_path, f.read()), ['pigar', 'pigar.tests'])

    def test_tar(self):
        tar_path = os.path.join(self._tmp_path, 'pigar-1.1.tar.gz')
        tarf = tarfile.open(tar_path, 'w:gz')
        fpath = os.path.join(os.path.dirname(__file__), 'fake_top_level.txt')
        try:
            tarf.add(fpath)
        finally:
            tarf.close()
        with open(tar_path, 'rb') as f:
            self.assertListEqual(
                top_level(tar_path, f.read()), ['pigar.tests'])



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import gzip
import tarfile
import zipfile
import re
import string
import io

from .utils import binary_type


class Archive(object):
    """Archive provides a consistent interface for unpacking
    compressed file.
    """

    def __init__(self, filename, fileobj):
        self._filename = filename
        self._fileobj = fileobj
        self._file = None
        self._names = None
        self._read = None

    @property
    def filename(self):
        return self._filename

    @property
    def names(self):
        """If name list is not required, do not get it."""
        if self._file is None:
            self._prepare()
        if not hasattr(self, '_namelist'):
            self._namelist = self._names()
        return self._namelist

    def close(self):
        """Close file object."""
        if self._file is not None:
            self._file.close()
        if hasattr(self, '_namelist'):
            del self._namelist
        self._filename = self._fileobj = None
        self._file = self._names = self._read = None

    def read(self, filename):
        """Read one file from archive."""
        if self._file is None:
            self._prepare()
        return self._read(filename)

    def unpack(self, to_path):
        """Unpack compressed files to path."""
        if self._file is None:
            self._prepare()
        self._safe_extractall(to_path)

    def _prepare(self):
        if self._filename.endswith(('.tar.gz', '.tar.bz2', '.tar.xz')):
            self._prepare_tarball()
        # An .egg file is actually just a .zip file
        # with a different extension, .whl too.
        elif self._filename.endswith(('.zip', '.egg', '.whl')):
            self._prepare_zip()
        else:
            raise ValueError("unreadable: {0}".format(self._filename))

    def _safe_extractall(self, to_path='.'):
        unsafe = []
        for name in self.names:
            if not self.is_safe(name):
                unsafe.append(name)
        if unsafe:
            raise ValueError("unsafe to unpack: {}".format(unsafe))
        self._file.extractall(to_path)

    def _prepare_zip(self):
        self._file = zipfile.ZipFile(self._fileobj)
        self._names = self._file.namelist
        self._read = self._file.read

    def _prepare_tarball(self):
        # tarfile has no read method
        def _read(filename):
            f = self._file.extractfile(filename)
            return f.read()

        self._file = tarfile.open(mode='r:*', fileobj=self._fileobj)
        self._names = self._file.getnames
        self._read = _read

    def is_safe(self, filename):
        return not (filename.startswith(("/", "\\")) or
                    (len(filename) > 1 and filename[1] == ":" and
                    filename[0] in string.ascii_letter) or
                    re.search(r"[.][.][/\\]", filename))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def top_level(url, data):
    """Read top level names from compressed file."""
    sb = io.BytesIO(data)
    txt = None
    with Archive(url, sb) as archive:
        file = None
        for name in archive.names:
            if name.lower().endswith('top_level.txt'):
                file = name
                break
        if file:
            txt = archive.read(file).decode('utf-8')
    sb.close()
    return [name.replace('/', '.') for name in txt.splitlines()] if txt else []



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import os
import re
import sys
import difflib

try:
    import colorama
except ImportError:
    colorama = None


PY32 = sys.version_info[:2] == (3, 2)

if sys.version_info[0] == 3:
    binary_type = bytes
else:
    binary_type = str


class Dict(dict):
    """Convert dict key object to attribute."""

    def __init__(self, *args, **kwargs):
        super(Dict, self).__init__(*args, **kwargs)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError('"{0}"'.format(name))

    def __setattr__(self, name, value):
        self[name] = value


# Color functions, win8 ...
_NONE = lambda text: text  # noqa
if colorama and not sys.platform.startswith('win'):
    _GREEN = lambda text: colorama.Fore.GREEN + text + colorama.Fore.RESET  # noqa
    _YELLOW = lambda text: colorama.Fore.YELLOW + text + colorama.Fore.RESET  # noqa
    _RED = lambda text: colorama.Fore.RED + text + colorama.Fore.RESET  # noqa
    _BLUE = lambda text: colorama.Fore.BLUE + text + colorama.Fore.RESET  # noqa
    _WHITE = lambda text: colorama.Fore.WHITE + text + colorama.Fore.RESET  # noqa
else:
    _GREEN = _YELLOW = _RED = _BLUE = _WHITE = _NONE

Color = Dict(
    GREEN=_GREEN, YELLOW=_YELLOW, RED=_RED,
    BLUE=_BLUE, WHITE=_WHITE, NONE=_NONE,
)


def print_table(rows, headers=['PACKAGE', 'CURRENT', 'LATEST']):
    """Print table. Such as:
     PACKAGE | CURRENT | LATEST
     --------+---------+-------
     pigar   | 0.4.5   | 0.5.0
    """
    end = len(headers) - 1
    hlens = [len(col) for col in headers]
    col_lens = hlens[:]
    for row in rows:
        col_lens = [max(col_lens[idx], len(col))
                    for idx, col in enumerate(row)]
    width = sum(col_lens) + end * 3 + 2
    print(' ' + '=' * width, end='\n ')
    for idx, header in enumerate(headers):
        print(" {0}{1}".format(header, (col_lens[idx] - hlens[idx]) * ' '),
              end=' |' if idx != end else '\n  ')
    for idx, col_len in enumerate(col_lens):
        print('{0}'.format(col_len * '-'), end='-+-' if idx != end else '\n ')
    for row in rows:
        for idx, col in enumerate(row):
            print(' {0}{1}'.format(col, (col_lens[idx] - len(col)) * ' '),
                  end=' |' if idx != end else '\n ')
    print('=' * width)


def parse_reqs(fpath):
    pkg_v_re = re.compile(r'^(?P<pkg>[^><==]+)[><==]{,2}(?P<version>.*)$')
    """Parse requirements file."""
    reqs = dict()
    with open(fpath, 'r') as f:
        for line in f:
            if line.startswith('#'):
                continue
            m = pkg_v_re.match(line.strip())
            if m:
                d = m.groupdict()
                reqs[d['pkg'].strip()] = d['version'].strip()
    return reqs


def cmp_to_key(cmp_func):
    """Convert a cmp=fcuntion into a key=function."""
    class K(object):
        def __init__(self, obj, *args):
            self.obj = obj

        def __lt__(self, other):
            return cmp_func(self.obj, other.obj) < 0

        def __gt__(self, other):
            return cmp_func(self.obj, other.obj) > 0

        def __eq__(self, other):
            return cmp_func(self.obj, other.obj) == 0

    return K


def compare_version(version1, version2):
    """Compare version number, such as 1.1.1 and 1.1b2.0."""
    v1, v2 = list(), list()

    for item in version1.split('.'):
        if item.isdigit():
            v1.append(int(item))
        else:
            v1.extend([i for i in _group_alnum(item)])
    for item in version2.split('.'):
        if item.isdigit():
            v2.append(int(item))
        else:
            v2.extend([i for i in _group_alnum(item)])

    while v1 and v2:
        item1, item2 = v1.pop(0), v2.pop(0)
        if item1 > item2:
            return 1
        elif item1 < item2:
            return -1

    if v1:
        return 1
    elif v2:
        return -1
    return 0


def _group_alnum(s):
    tmp = list()
    flag = 1 if s[0].isdigit() else 0
    for c in s:
        if c.isdigit():
            if flag == 0:
                yield ''.join(tmp)
                tmp = list()
                flag = 1
            tmp.append(c)
        elif c.isalpha():
            if flag == 1:
                yield int(''.join(tmp))
                tmp = list()
                flag = 0
            tmp.append(c)
    last = ''.join(tmp)
    yield (int(last) if flag else last)


def parse_git_config(path):
    """Parse git config file."""
    config = dict()
    section = None

    with open(os.path.join(path, 'config'), 'r') as f:
        for line in f:
            line = line.strip()
            if line.startswith('['):
                section = line[1: -1].strip()
                config[section] = dict()
            elif section:
                key, value = line.replace(' ', '').split('=')
                config[section][key] = value
    return config


def lines_diff(lines1, lines2):
    """Show difference between lines."""
    is_diff = False
    diffs = list()

    for line in difflib.ndiff(lines1, lines2):
        if not is_diff and line[0] in ('+', '-'):
            is_diff = True
        diffs.append(line)

    return is_diff, diffs



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import os
import sys
import unittest

from ..__main__ import GenerateReqs
from ..reqs import is_stdlib


class ReqsTests(unittest.TestCase):

    def setUp(self):
        self._installed_packages = {
            'foo': ('Foo', '0.1.0'),
            'bar': ('Bar', '1.1.1'),
            'baz': ('Baz', '2.2.2'),
            'foobaz': ('FooBaz', '20151110'),
            'mod': ('Mod', '1.0.0'),
            'name': ('Name', '1.0.0'),
            'pkg': ('Pkg', '1.0.0'),
        }
        self._module_infos = {
            'os': ['pigar/tests/imports_example/example1.py: 3'],
            'sys': ['pigar/tests/imports_example/example1.py: 4'],
            'collections': ['pigar/tests/imports_example/example1.py: 5'],
            'foobar': ['pigar/tests/imports_example/example1.py: 7'],
            'FooBar': ['pigar/tests/imports_example/example1.py: 7'],
            'Foo': ['pigar/tests/imports_example/example1.py: 10'],
            'Bar': ['pigar/tests/imports_example/example1.py: 13'],
            'FooBaz': ['pigar/tests/imports_example/example1.py: 16'],
            'json': ['pigar/tests/imports_example/example1.py: 20'],
            'itertools': ['pigar/tests/imports_example/example1.py: 26'],
            'Baz': ['pigar/tests/imports_example/example1.py: 30'],
            'Queue': ['pigar/tests/imports_example/example1.py: 36'],
            'bisect': ['pigar/tests/imports_example/example1.py: 38'],
            'Mod': ['pigar/tests/imports_example/example1.py: 44'],
            'Name': ['pigar/tests/imports_example/example1.py: 46'],
            'Pkg': ['pigar/tests/imports_example/example1.py: 48'],
            'urlparse': ['pigar/tests/imports_example/example2.py: 4'],
            'urllib': ['pigar/tests/imports_example/example2.py: 6'],
            '__builtin__': ['pigar/tests/imports_example/example2.py: 8'],
            'builtins': ['pigar/tests/imports_example/example2.py: 10'],
            'example1': ['pigar/tests/imports_example/example2.py: 12'],
        }
        self._try_imports = ['urlparse', 'urllib', '__builtin__', 'builtins']
        if sys.version_info[0] == 2:
            self._installed_packages.update({'foobar': ('FooBar', '3.6.9')})
        self._path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), 'imports_example/'))

    def tearDown(self):
        del self._installed_packages
        del self._path
        del self._module_infos

    @unittest.skipIf(sys.version_info[0] != 3, 'Not python 3.x')
    def test_py3_reqs(self):
        pv = {k: v for k, v in self._installed_packages.values()}
        reqs, try_imports, guess = self._extract_reqs()

        self.assertListEqual(sorted(reqs.keys()), sorted(pv.keys()))
        self.assertListEqual(sorted(try_imports), sorted(self._try_imports))
        # Assume 'foobar' is Py3 builtin package, no need install.
        self.assertListEqual(
            sorted(guess.keys()),
            sorted(['Queue', '__builtin__', 'foobar', 'urlparse', 'example1'])
        )
        self._check_detail(reqs, pv)
        self._check_detail(guess, pv, False)

    @unittest.skipIf(sys.version_info[0] != 2, 'Not python 2.x')
    def test_py2_reqs(self):
        self._installed_packages.update({'foobar': ('FooBar', '3.3.3')})
        pv = {k: v for k, v in self._installed_packages.values()}
        reqs, try_imports, guess = self._extract_reqs()

        self.assertListEqual(sorted(try_imports), sorted(self._try_imports))
        self.assertListEqual(sorted(reqs.keys()), sorted(pv.keys()))
        self.assertListEqual(guess.keys(), ['builtins', 'example1'])
        self._check_detail(reqs, pv)
        self._check_detail(guess, pv, False)

    def _check_detail(self, reqs_mods, pv, version=True):
        for pkg, detail in reqs_mods.sorted_items():
            if version:
                if pkg not in pv:
                    self.fail('"{0}" not installed'.format(pkg))
                self.assertEqual(detail.version, pv[pkg])
            self.assertListEqual(
                sorted(detail.comments.sorted_items()),
                sorted(self._module_infos[pkg])
            )

    def _extract_reqs(self):
        gr = GenerateReqs('', self._path, [], self._installed_packages)
        return gr.extract_reqs()


class StdlibTest(unittest.TestCase):

    def test_stdlib(self):
        self.assertTrue(is_stdlib('os'))
        self.assertTrue(is_stdlib('sys'))



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import os
import argparse

from ._version import version


def parse_args(args=None):
    parser = argparse.ArgumentParser(
        prog='pigar',
        #  usage='%(prog)s [options]',
        description='Python requirements tool -- %(prog)s, it will '
        'do only one thing at each time. Default action is generate'
        ' requirements.txt in current directory.'
    )
    parser.add_argument(
        '-v', '--version',
        action='version',
        version='%(prog)s {0}'.format(version),
        help='show %(prog)s version information and exit')
    parser.add_argument(
        '-u', '--update',
        dest='update_db',
        action='store_true',
        help='update database, use it when %(prog)s failed you, exit when'
        ' action done')
    parser.add_argument(
        '-s',  # '--search',
        dest='search_names',
        nargs='+',
        metavar='NAME',
        default=[],
        help='search package name by import name, use it if you do not '
        'know import name come from which package, exit when action done')
    parser.add_argument(
        '-c',  # '--check',
        dest='check_path',
        metavar='PATH',
        nargs='?',
        type=path_check,
        const=os.getcwd(),
        help='check requirements for the latest version. If file path not '
        'given, search *requirements.txt in current directory, if not found, '
        'generate file requirements.txt, exit when action done')
    parser.add_argument(
        '-l',  # '--log_level'
        dest='log_level',
        nargs=1,
        type=log_level_check,
        default=['error'],
        help='show given level log messages, argument can be '
        '(ERROR, WARNING, INFO, DEBUG), case-insensitive')
    parser.add_argument(
        '-i',  # '--ignore'
        dest='ignores',
        nargs='+',
        metavar='DIR',
        default=[],
        type=ignore_dirs_check,
        help='given a list of directory to ignore, relative directory, '
        '*used for* -c and default action')
    parser.add_argument(
        '-p',  # '--path',
        dest='save_path',
        nargs=1,
        type=path_check,
        default=[os.path.join(os.getcwd(), 'requirements.txt')],
        help='save requirements in given file path, *used for* default action')
    parser.add_argument(
        '-P',  # '--projectpath',
        dest='project_path',
        nargs=1,
        type=projectpath_check,
        default=[os.getcwd()],
        help='project path, which is directory, *used for* default action')
    parser.add_argument(
        '-o',  # '--comparison-operator',
        dest='comparison_operator',
        nargs=1,
        type=comparison_operator_check,
        default=['=='],
        help='The comparison operator for versions, alternatives: [==, ~=, >=]'
    )
    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args=args)
    return (args.log_level[0], args.update_db, args.check_path,
            args.search_names, args.ignores, args.save_path[0],
            args.project_path[0], args.comparison_operator[0])


def log_level_check(level):
    levels = ('ERROR', 'WARNING', 'INFO', 'DEBUG')
    if level.upper() in levels:
        return level
    raise argparse.ArgumentTypeError(
        '"{0}" not in "{1}"'.format(level, ', '.join(levels)))


def ignore_dirs_check(d):
    ignore = os.path.abspath(d)
    if not os.path.isdir(ignore):
        raise argparse.ArgumentTypeError('"{0}" is not directory.'.format(d))
    return ignore


def path_check(path):
    path = os.path.abspath(path)
    fdir = os.path.dirname(path)
    if os.path.isdir(fdir):
        return path
    msg = '"{0}" is not a valid file path.'.format(path)
    raise argparse.ArgumentTypeError(msg)


def projectpath_check(path):
    path = os.path.abspath(path)
    if os.path.isdir(path):
        return path
    msg = '"{0}" is not a valid directory path.'.format(path)
    raise argparse.ArgumentTypeError(msg)


def comparison_operator_check(op, supported_ops=('==', '~=', '>=')):
    if op in supported_ops:
        return op
    msg = 'invalid operator: {0}, supported operators: {1}'.format(
        op, supported_ops)
    raise argparse.ArgumentTypeError(msg)



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import os
import sys
import fnmatch
import importlib
import imp
import ast
import doctest
import collections
import functools
try:
    from types import FileType  # py2
except ImportError:
    from io import IOBase as FileType  # py3

from .log import logger
from .utils import parse_git_config
from .modules import ImportedModules


def project_import_modules(project_path, ignores):
    """Get entire project all imported modules."""
    modules = ImportedModules()
    try_imports = set()
    local_mods = list()
    cur_dir = os.getcwd()
    ignore_paths = collections.defaultdict(set)
    if not ignores:
        ignore_paths[project_path].add('.git')
    else:
        for path in ignores:
            parent_dir = os.path.dirname(path)
            ignore_paths[parent_dir].add(os.path.basename(path))

    logger.info('Extracting project: {0}'.format(project_path))
    for dirpath, dirnames, files in os.walk(project_path, followlinks=True):
        if dirpath in ignore_paths:
            dirnames[:] = [d for d in dirnames
                           if d not in ignore_paths[dirpath]]
        logger.info('Extracting directory: {0}'.format(dirpath))
        py_files = list()
        for fn in files:
            # C extension.
            if fn.endswith('.so'):
                local_mods.append(fn[:-3])
            # Normal Python file.
            if fn.endswith('.py'):
                local_mods.append(fn[:-3])
                py_files.append(fn)
        if '__init__.py' in files:
            local_mods.append(os.path.basename(dirpath))
        for file in py_files:
            fpath = os.path.join(dirpath, file)
            fake_path = fpath.split(cur_dir)[1][1:]
            logger.info('Extracting file: {0}'.format(fpath))
            with open(fpath, 'rb') as f:
                fmodules, try_ipts = file_import_modules(fake_path, f.read())
                modules |= fmodules
                try_imports |= try_ipts

    logger.info('Finish extracting in project: {0}'.format(project_path))
    return modules, try_imports, local_mods


def file_import_modules(fpath, fdata):
    """Get single file all imported modules."""
    modules = ImportedModules()
    str_codes = collections.deque([(fdata, 1)])
    try_imports = set()

    while str_codes:
        str_code, lineno = str_codes.popleft()
        ic = ImportChecker(fpath, lineno)
        try:
            parsed = ast.parse(str_code)
            ic.visit(parsed)
        # Ignore SyntaxError in Python code.
        except SyntaxError:
            pass
        modules |= ic.modules
        str_codes.extend(ic.str_codes)
        try_imports |= ic.try_imports
        del ic

    return modules, try_imports


class ImportChecker(object):

    def __init__(self, fpath, lineno):
        self._fpath = fpath
        self._lineno = lineno - 1
        self._modules = ImportedModules()
        self._str_codes = collections.deque()
        self._try_imports = set()

    def visit_Import(self, node, try_=False):
        """As we know: `import a [as b]`."""
        lineno = node.lineno + self._lineno
        for alias in node.names:
            self._modules.add(alias.name, self._fpath, lineno)
            if try_:
                self._try_imports.add(alias.name)

    def visit_ImportFrom(self, node, try_=False):
        """
        As we know: `from a import b [as c]`. If node.level is not 0,
        import statement like this `from .a import b`.
        """
        mod_name = node.module
        level = node.level
        if mod_name is None:
            level -= 1
            mod_name = ''
        for alias in node.names:
            name = level*'.' + mod_name + '.' + alias.name
            self._modules.add(name, self._fpath, node.lineno + self._lineno)
            if try_:
                self._try_imports.add(name)

    def visit_TryExcept(self, node):
        """
        If modules which imported by `try except` and not found,
        maybe them come from other Python version.
        """
        for ipt in node.body:
            if ipt.__class__.__name__.startswith('Import'):
                method = 'visit_' + ipt.__class__.__name__
                getattr(self, method)(ipt, True)
        for handler in node.handlers:
            for ipt in handler.body:
                if ipt.__class__.__name__.startswith('Import'):
                    method = 'visit_' + ipt.__class__.__name__
                    getattr(self, method)(ipt, True)

    # For Python 3.3+
    visit_Try = visit_TryExcept

    def visit_Exec(self, node):
        """
        Check `expression` of `exec(expression[, globals[, locals]])`.
        **Just available in python 2.**
        """
        if hasattr(node.body, 's'):
            self._str_codes.append((node.body.s, node.lineno + self._lineno))
        # PR#13: https://github.com/damnever/pigar/pull/13
        # Sometimes exec statement may be called with tuple in Py2.7.6
        elif hasattr(node.body, 'elts') and len(node.body.elts) >= 1:
            self._str_codes.append(
                (node.body.elts[0].s, node.lineno + self._lineno))

    def visit_Expr(self, node):
        """
        Check `expression` of `eval(expression[, globals[, locals]])`.
        Check `expression` of `exec(expression[, globals[, locals]])`
        in python 3.
        Check `name` of `__import__(name[, globals[, locals[,
        fromlist[, level]]]])`.
        Check `name` or `package` of `importlib.import_module(name,
        package=None)`.
        """
        # Built-in functions
        value = node.value
        if isinstance(value, ast.Call):
            if hasattr(value.func, 'id'):
                if (value.func.id == 'eval' and
                        hasattr(node.value.args[0], 's')):
                    self._str_codes.append(
                        (node.value.args[0].s, node.lineno + self._lineno))
                # **`exec` function in Python 3.**
                elif (value.func.id == 'exec' and
                        hasattr(node.value.args[0], 's')):
                    self._str_codes.append(
                        (node.value.args[0].s, node.lineno + self._lineno))
                # `__import__` function.
                elif (value.func.id == '__import__' and
                        len(node.value.args) > 0 and
                        hasattr(node.value.args[0], 's')):
                    self._modules.add(node.value.args[0].s, self._fpath,
                                      node.lineno + self._lineno)
            # `import_module` function.
            elif getattr(value.func, 'attr', '') == 'import_module':
                module = getattr(value.func, 'value', None)
                if (module is not None and
                        getattr(module, 'id', '') == 'importlib'):
                    args = node.value.args
                    arg_len = len(args)
                    if arg_len > 0 and hasattr(args[0], 's'):
                        name = args[0].s
                        if not name.startswith('.'):
                            self._modules.add(name, self._fpath,
                                              node.lineno + self._lineno)
                        elif arg_len == 2 and hasattr(args[1], 's'):
                            self._modules.add(args[1].s, self._fpath,
                                              node.lineno + self._lineno)

    def visit_FunctionDef(self, node):
        """
        Check docstring of function, if docstring is used for doctest.
        """
        docstring = self._parse_docstring(node)
        if docstring:
            self._str_codes.append((docstring, node.lineno + self._lineno + 2))

    def visit_ClassDef(self, node):
        """
        Check docstring of class, if docstring is used for doctest.
        """
        docstring = self._parse_docstring(node)
        if docstring:
            self._str_codes.append((docstring, node.lineno + self._lineno + 2))

    def visit(self, node):
        """Visit a node, no recursively."""
        for node in ast.walk(node):
            method = 'visit_' + node.__class__.__name__
            getattr(self, method, lambda x: x)(node)

    @staticmethod
    def _parse_docstring(node):
        """Extract code from docstring."""
        docstring = ast.get_docstring(node)
        if docstring:
            parser = doctest.DocTestParser()
            try:
                dt = parser.get_doctest(docstring, {}, None, None, None)
            except ValueError:
                # >>> 'abc'
                pass
            else:
                examples = dt.examples
                return '\n'.join([example.source for example in examples])
        return None

    @property
    def modules(self):
        return self._modules

    @property
    def str_codes(self):
        return self._str_codes

    @property
    def try_imports(self):
        return set((name.split('.')[0] if name and '.' in name else name)
                   for name in self._try_imports)


def _checked_cache(func):
    checked = dict()

    @functools.wraps(func)
    def _wrapper(name):
        if name not in checked:
            checked[name] = func(name)
        return checked[name]

    return _wrapper


@_checked_cache
def is_stdlib(name):
    """Check whether it is stdlib module."""
    exist = True
    module_info = ('', '', '')
    try:
        module_info = imp.find_module(name)
    except ImportError:
        try:
            # __import__(name)
            importlib.import_module(name)
            module_info = imp.find_module(name)
            sys.modules.pop(name)
        except ImportError:
            exist = False
    # Testcase: ResourceWarning
    if isinstance(module_info[0], FileType):
        module_info[0].close()
    mpath = module_info[1]
    if exist and (mpath is not None and
                  ('site-packages' in mpath or
                   'dist-packages' in mpath or
                   ('bin/' in mpath and mpath.endswith('.py')))):
        exist = False
    return exist


def get_installed_pkgs_detail():
    """Get mapping for import top level name
    and install package name with version.
    """
    mapping = dict()

    for path in sys.path:
        if os.path.isdir(path) and path.rstrip('/').endswith(
                ('site-packages', 'dist-packages')):
            mapping.update(_search_path(path))

    return mapping


def _search_path(path):
    mapping = dict()

    for file in os.listdir(path):
        # Install from PYPI.
        if fnmatch.fnmatch(file, '*-info'):
            top_level = os.path.join(path, file, 'top_level.txt')
            pkg_name, version = file.split('-')[:2]
            if version.endswith('dist'):
                version = version.rsplit('.', 1)[0]
            # Issue for ubuntu: sudo pip install xxx
            elif version.endswith('egg'):
                version = version.rsplit('.', 1)[0]
            mapping[pkg_name] = (pkg_name, version)
            if not os.path.isfile(top_level):
                continue
            with open(top_level, 'r') as f:
                for line in f:
                    mapping[line.strip()] = (pkg_name, version)

        # Install from local and available in GitHub.
        elif fnmatch.fnmatch(file, '*-link'):
            link = os.path.join(path, file)
            if not os.path.isfile(link):
                continue
            # Link path.
            with open(link, 'r') as f:
                for line in f:
                    line = line.strip()
                    if line != '.':
                        dev_dir = line
            if not dev_dir:
                continue
            # Egg info path.
            info_dir = [_file for _file in os.listdir(dev_dir)
                        if _file.endswith('egg-info')]
            if not info_dir:
                continue
            info_dir = info_dir[0]
            top_level = os.path.join(dev_dir, info_dir, 'top_level.txt')
            # Check whether it can be imported.
            if not os.path.isfile(top_level):
                continue

            # Check .git dir.
            git_path = os.path.join(dev_dir, '.git')
            if os.path.isdir(git_path):
                config = parse_git_config(git_path)
                url = config.get('remote "origin"', {}).get('url')
                if not url:
                    continue
                branch = 'branch "master"'
                if branch not in config:
                    for section in config:
                        if 'branch' in section:
                            branch = section
                            break
                if not branch:
                    continue
                branch = branch.split()[1][1:-1]

                pkg_name = info_dir.split('.egg')[0]
                git_url = 'git+{0}@{1}#egg={2}'.format(url, branch, pkg_name)
                with open(top_level, 'r') as f:
                    for line in f:
                        mapping[line.strip()] = ('-e', git_url)
    return mapping



# -*- coding: utf-8 -*-

import unittest

from ..modules import ImportedModules, ReqsModules, _Locations


class LocationsTests(unittest.TestCase):

    def setUp(self):
        self._data = {
            'oo/xx.py': 33,
            'bar/baz.py': 10,
        }

    def test_add(self):
        loc = _Locations()
        for file, lineno in self._data.items():
            loc.add(file, lineno)

        for file in self._data:
            if file not in loc:
                self.fail('add "{0}" failed'.format(file))
            else:
                self.assertEqual(loc[file], [self._data[file]])

        loc.add('oo/xx.py', 2)
        self.assertListEqual(sorted(loc['oo/xx.py']), [2, 33])

    def test_extend(self):
        loc1 = _Locations()
        loc2 = _Locations()
        r1, r2 = self._data.items()
        loc1.add(r1[0], r1[1])
        loc2.add(r2[0], r2[1])

        loc1.extend(loc2)

        self.assertListEqual(sorted(loc1.items()),
                             sorted([(k, [v]) for k, v in self._data.items()]))

    def test_sorted_items(self):
        loc = _Locations()
        for file, lineno in self._data.items():
            loc.add(file, lineno)

        target = ['{0}: {1}'.format(k, v) for k, v in self._data.items()]
        self.assertListEqual(loc.sorted_items(), sorted(target))


class ImportedModulesTests(unittest.TestCase):

    def setUp(self):
        self._mapping = {
            'pigar': ('oo/xx.py', 33),
            'foo': ('bar/baz.py', 10),
        }

    def test_add(self):
        im = ImportedModules()
        for name, (file, lineno) in self._mapping.items():
            im.add(name, file, lineno)

        self.assertListEqual(sorted(im.keys()), sorted(self._mapping.keys()))
        for name in self._mapping:
            if name not in im:
                self.fail('add "{0}" failed!'.format(name))
            else:
                file, linenos = list(im[name].items())[0]
                self.assertTupleEqual((file, linenos[0]), self._mapping[name])

        im.add('pigar', 'oo/xx.py', 2)
        file, linenos = list(im['pigar'].items())[0]
        self.assertListEqual(sorted(linenos), [2, 33])

    def test_or(self):
        im1 = ImportedModules()
        im2 = ImportedModules()
        r1, r2 = list(self._mapping.items())
        im1.add(r1[0], r1[1][0], r1[1][1])
        im2.add(r2[0], r2[1][0], r2[1][1])

        im1 |= im2
        self.assertListEqual(sorted(im1.keys()), sorted(self._mapping.keys()))


class ReqsModulesTest(unittest.TestCase):

    def setUp(self):
        loc1 = _Locations()
        loc2 = _Locations()
        loc1.add('oo/xx.py', 33)
        loc2.add('bar/baz.py', 10)
        self._data = {
            'pigar': ('9.9.9', loc1),
            'foo': ('3.3.3', loc2),
        }

    def test_add(self):
        rm = ReqsModules()
        for pkg, (ver, loc) in self._data.items():
            rm.add(pkg, ver, loc)

        for pkg in self._data:
            if pkg not in rm:
                self.fail('add "{0}" failed'.format(pkg))
            else:
                detail = rm[pkg]
                val = self._data[pkg]
                self.assertEqual(detail.version, val[0])
                self.assertIs(detail.comments, val[1])

        loc = _Locations()
        loc.add('foobar.py', 2)
        rm.add('pigar', '9.9.9', loc)
        self.assertListEqual(rm['pigar'].comments.sorted_items(),
                             sorted(['oo/xx.py: 33', 'foobar.py: 2']))



# -*- coding: utf-8 -*-

try:
    import urlparse as parse  # Py2
except ImportError:
    from urllib import parse  # Py3
try:
    import __builtin__ as builtins  # Py2
except ImportError:
    import builtins  # Py3

import example1



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import gzip
import json
import collections
import io
import threading
import multiprocessing
from multiprocessing import Queue as ProcessSharableQueue
try:  # py2
    from HTMLParser import HTMLParser
    from urlparse import urljoin
    from Queue import Empty, Queue as ThreadSharableQueue
except ImportError:  # py3
    from html.parser import HTMLParser
    from urllib.parse import urljoin
    from queue import Empty, Queue as ThreadSharableQueue

import concurrent.futures
import requests

from .db import database, Database
from .unpack import top_level
from .log import logger
from .utils import Color, compare_version, cmp_to_key, binary_type


PYPI_URL = 'https://pypi.org'
PKG_URL = urljoin(PYPI_URL, '/pypi/{0}')
PKGS_URL = urljoin(PYPI_URL, '/simple/')
PKG_INFO_URL = urljoin(PYPI_URL, '/pypi/{0}/json')
ACCEPTABLE_EXT = ('.whl', '.egg', '.tar.gz', '.tar.bz2', '.zip')


def search_names(names, installed_pkgs):
    """Search package information by names(`import XXX`).
    """
    dler = Downloader()
    results = collections.defaultdict(list)
    not_found = list()

    for name in names:
        logger.info('Searching package name for "{0}" ...'.format(name))
        # If exists in local environment, do not check on the PyPI.
        if name in installed_pkgs:
            results[name].append(list(installed_pkgs[name]) + ['local'])
        # Check information on the PyPI.
        else:
            rows = None
            with database() as db:
                rows = db.query_all(name)
            if rows:
                for row in rows:
                    version = dler.download_package(row.package).version()
                    results[name].append((row.package, version, 'PyPI'))
            else:
                not_found.append(name)
    return results, not_found


def check_latest_version(package):
    """Check package latest version in PyPI."""
    return Downloader().download_package(package).version()


def update_db():
    """Update database."""
    print(Color.BLUE('Starting update database ...'))
    print(Color.YELLOW('The process will take a long time!!!'))
    logger.info('Crawling "{0}" ...'.format(PKGS_URL))
    try:
        updater = Updater()
    except Exception:
        logger.error("Fail to fetch all packages: ", exc_info=True)
        print(Color.RED('Operation aborted'))
        return

    try:
        updater.run()
        updater.wait()
    except (KeyboardInterrupt, SystemExit):
        # FIXME(damnever): the fucking signal..
        updater.cancel()
        print(Color.BLUE('Operation canceled!'))
    else:
        print(Color.BLUE('Operation done!'))


class Updater(object):
    _CPU_NUM = multiprocessing.cpu_count()  # C*2+1? The bandwidth matters..

    def __init__(self, proc_num=_CPU_NUM):
        self._proc_num = proc_num

        downloader = Downloader()
        index = downloader.download_index()
        downloader.close()

        threads_total = self._proc_num * 6
        if threads_total < 24:
            threads_total = 24
        self._threads_per_proc = threads_total // self._proc_num

        # XXX(damnever): using the magic __new__???
        self._thread_updater = None
        if proc_num == 1:
            pkg_names = ThreadSharableQueue()
            _extract_pkg_names(index, pkg_names.put)
            self._thread_updater = ThreadPoolUpdater(
                pkg_names, threads_total)
        else:
            self._pkg_names = ProcessSharableQueue()
            t = threading.Thread(
                target=_extract_pkg_names,
                args=(index, self._pkg_names.put)
            )
            t.daemon = True
            t.start()
            self._feed_thread = t
            self._procs = []

        with database():
            pass

    def run(self):
        if self._thread_updater is not None:
            self._thread_updater.run()
            return

        for _ in range(self._proc_num):
            proc = multiprocessing.Process(
                target=self._proc_main,
                args=(self._pkg_names, self._threads_per_proc),
            )
            proc.daemon = True
            proc.start()
            self._procs.append(proc)

    def wait(self):
        if self._thread_updater is not None:
            return self._thread_updater.wait()

        self._feed_thread.join()
        [proc.join() for proc in self._procs]

    def cancel(self):
        if self._thread_updater is not None:
            return self._thread_updater.cancel()

        [proc.terminate() for proc in self._procs]
        [proc.join(timeout=1) for proc in self._procs]

    def _proc_main(self, pkg_names, workernum):
        tupdater = ThreadPoolUpdater(pkg_names, workernum)
        tupdater.run()
        tupdater.wait()


class ThreadPoolUpdater(object):
    def __init__(self, pkg_names, workernum=24):
        self._max_workers = workernum
        self._pkg_names = pkg_names
        self._futures = []

    def run(self):
        with concurrent.futures.ThreadPoolExecutor(
                max_workers=self._max_workers) as executor:
            # Incase of unexpected error happens.
            for _ in range(self._max_workers*3):
                future = executor.submit(self.extract_and_update)
                self._futures.append(future)

    def wait(self):
        for future in concurrent.futures.as_completed(self._futures):
            try:
                error = future.exception()
            except concurrent.futures.CancelledError:
                break
            if error is not None:
                logger.error('Unexpected error: {}'.format(error))

    def cancel(self):
        for future in self._futures:
            future.cancel()

    def extract_and_update(self):
        dler = Downloader()
        db = Database()
        try:
            while 1:
                try:
                    pkg_name = self._pkg_names.get(block=False)
                    logger.info('Processing package: %s', pkg_name)
                    pkg = dler.download_package(pkg_name)
                    top_levels = pkg.top_levels()
                    db.insert_package_with_imports(pkg_name, top_levels)
                except (requests.RequestException, KeyError) as e:
                    logger.error(
                        ('Maybe package "%s" is no longer available'
                         ' or it is non-standard: %r'), pkg_name, e)
        except Empty:
            pass
        except Exception:
            logger.debug('Thread exited:', exc_info=True)
            raise
        finally:
            dler.close()
            db.close()


class Downloader(object):
    _HEADERS = {
        'Accept': '*/*',
        'Accept-Encoding': 'gzip',
        'Accept-Language': 'en-US,en;q=0.8,zh-CN;q=0.6,zh;q=0.4',
        'User-Agent': ('Mozilla/5.0 (X11; Linux x86_64; rv:13.0) '
                       'AppleWebKit/537.36 (KHTML, like Gecko) '
                       'Chrome/44.0.2403.157 Safari/537.36'),
    }

    def __init__(self):
        self._session = requests.Session()

    def download(self, url, try_decode=True):
        # XXX(damnever): timeout?
        resp = self._session.get(url, headers=self._HEADERS)
        resp.raise_for_status()
        data = resp.content
        if 'gzip' == resp.headers.get('Content-Encoding'):
            try:
                with gzip.GzipFile(fileobj=io.BytesIO(data)) as gz:
                    data = gz.read()
            except (OSError, IOError):  # Not a gzip file
                pass
        if try_decode and isinstance(data, binary_type):
            data = data.decode('utf-8')
        return data

    def download_index(self):
        return self.download(PKGS_URL)

    def download_package(self, name):
        pkg_info = self.download(PKG_INFO_URL.format(name))
        return Package(name, json.loads(pkg_info), self)

    def close(self):
        self._session.close()


class Package(object):
    def __init__(self, name, pkg_info, downloader):
        self._name = name
        self._pkg_info = pkg_info
        self._downloader = downloader

    def version(self):
        info = self._pkg_info
        try:
            latest = info['info'].get('version', None)
            if latest is not None:
                return latest
            latest = sorted(info['releases'], key=cmp_to_key(compare_version))
            latest = latest[-1]
            return latest
        except KeyError:
            return 'unknown'

    def top_levels(self):
        # Extracting names which can be imported.
        url = None
        for item in self._pkg_info['urls']:
            if item['filename'].endswith(ACCEPTABLE_EXT):
                url = item['url']
                break
        if url is None:
            return []
        pkg = self._downloader.download(url, try_decode=False)
        try:
            return top_level(url, pkg)
        except Exception:
            return []


def _extract_pkg_names(html, put):
    """Extract data from html."""
    class PackageNameParser(HTMLParser):
        def handle_starttag(self, tag, attrs):
            if tag == 'a':
                attrs = dict(attrs)
                if attrs.get('href', None):
                    name = attrs['href'].strip('/').split('/')[-1]
                    put(name)

    PackageNameParser().feed(html)



# -*- coding: utf-8 -*-
# Reference:
# http://stackoverflow.com/questions/16571150/how-to-capture-stdout-output-from-a-python-function-call

from __future__ import print_function, division, absolute_import

import sys
try:
    from cStringIO import StringIO
except ImportError:
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO


class CaptureOutput(list):

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._strio = StringIO()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.extend(self._strio.getvalue().splitlines())
        self._strio.close()
        sys.stdout = self._stdout



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import collections


# FIXME: Just a workaround, not a radical cure..
_special_cases = {
    "dogpile.cache": "dogpile.cache",
    "dogpile.core": "dogpile.core",
    "ruamel.yaml": "ruamel.yaml",
    "ruamel.ordereddict": "ruamel.ordereddict",
}


class Modules(dict):
    """Modules object will be used to store modules information."""

    def __init__(self):
        super(Modules, self).__init__()


class ImportedModules(Modules):

    def __init__(self):
        super(ImportedModules, self).__init__()

    def add(self, name, file, lineno):
        if name is None:
            return

        names = list()
        special_name = '.'.join(name.split('.')[:2])
        # Flask extension.
        if name.startswith('flask.ext.'):
            names.append('flask')
            names.append('flask_' + name.split('.')[2])
        # Special cases..
        elif special_name in _special_cases:
            names.append(_special_cases[special_name])
        # Other.
        elif '.' in name and not name.startswith('.'):
            names.append(name.split('.')[0])
        else:
            names.append(name)

        for nm in names:
            if nm not in self:
                self[nm] = _Locations()
            self[nm].add(file, lineno)

    def __or__(self, obj):
        for name, locations in obj.items():
            for file, linenos in locations.items():
                for lineno in linenos:
                    self.add(name, file, lineno)
        return self


class ReqsModules(Modules):

    _Detail = collections.namedtuple('Detail', ['version', 'comments'])

    def __init__(self):
        super(ReqsModules, self).__init__()
        self._sorted = None

    def add(self, package, version, locations):
        if package in self:
            self[package].comments.extend(locations)
        else:
            self[package] = self._Detail(version, locations)

    def sorted_items(self):
        if self._sorted is None:
            self._sorted = sorted(self.items())
        return self._sorted

    def remove(self, *names):
        for name in names:
            if name in self:
                self.pop(name)
        self._sorted = None


class _Locations(dict):
    """_Locations store code locations(file, linenos)."""

    def __init__(self):
        super(_Locations, self).__init__()
        self._sorted = None

    def add(self, file, lineno):
        if file in self and lineno not in self[file]:
            self[file].append(lineno)
        else:
            self[file] = [lineno]

    def extend(self, obj):
        for file, linenos in obj.items():
            for lineno in linenos:
                self.add(file, lineno)

    def sorted_items(self):
        if self._sorted is None:
            self._sorted = [
                '{0}: {1}'.format(f, ','.join([str(n) for n in sorted(ls)]))
                for f, ls in sorted(self.items())
            ]
        return self._sorted



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import os
import sys
import fnmatch
import codecs

from .cmd import parse_args
from .reqs import project_import_modules, is_stdlib, get_installed_pkgs_detail
from .pypi import update_db, check_latest_version, search_names
from .db import database
from .utils import Color, parse_reqs, print_table, lines_diff
from .log import logger, enable_pretty_logging
from .modules import ReqsModules


class Main(object):

    def __init__(self):
        # Parse command arguments.
        (log_level, updatedb, check_path, names, ignores, save_path,
         project_path, comparison_operator) = parse_args()
        # Enable logging.
        enable_pretty_logging(log_level=log_level)
        # Just allow do one thing at each time.
        if updatedb:
            self.update_db()
        elif check_path:
            self.check_reqs_latest_version(
                check_path, ignores, comparison_operator)
        elif names:
            self.search_package_by_name(names)
        else:
            self.generate_reqs(save_path, project_path,
                               ignores, comparison_operator)

    @property
    def installed_pkgs(self):
        # Lazy calculation.
        if not hasattr(self, '_installed_pkgs'):
            self._installed_pkgs = get_installed_pkgs_detail()
        return self._installed_pkgs

    def update_db(self):
        """Update database."""
        update_db()

    def search_package_by_name(self, names):
        """Search package name by import name."""
        print(Color.BLUE('Starting search names ...'))
        found, not_found = search_names(names, self.installed_pkgs)
        for name in found:
            print('Found package(s) for "{0}":'.format(Color.GREEN(name)))
            print_table(found[name], headers=['PACKAGE', 'VERSION', 'WHERE'])
        if not_found:
            msg = '"{0}" not found.\n'.format(Color.RED(', '.join(not_found)))
            msg += 'Maybe you need update database.'
            print(Color.YELLOW(msg))

    def check_reqs_latest_version(self, check_path, ignores,
                                  comparison_operator):
        """Check requirements latest version."""
        print(Color.BLUE('Starting check requirements latest version ...'))
        files = list()
        reqs = dict()
        pkg_versions = list()
        # If no requirements file given, check in current directory.
        if os.path.isdir(check_path):
            print(Color.BLUE('Searching file in "{0}" ...'.format(check_path)))
            for fn in os.listdir(check_path):
                if fnmatch.fnmatch(fn, '*requirements.txt'):
                    files.append(os.path.abspath(fn))
            # If not found in directory, generate requirements.
            if not files:
                print(Color.YELLOW('Requirements file not found, '
                                   'generate requirements ...'))
                save_path = os.path.join(check_path, 'requirements.txt')
                self.generate_reqs(save_path, check_path,
                                   ignores, comparison_operator)
                files.append(save_path)
        else:
            files.append(check_path)
        for fpath in files:
            reqs.update(parse_reqs(fpath))

        print(Color.BLUE('Checking requirements latest version ...'))
        installed_pkgs = {v[0]: v[1] for k, v in self.installed_pkgs.items()}
        for pkg in reqs:
            current = reqs[pkg]
            # If no version specifies in requirements,
            # check in installed packages.
            if current == '' and pkg in installed_pkgs:
                current = installed_pkgs[pkg]
            logger.info('Checking "{0}" latest version ...'.format(pkg))
            latest = check_latest_version(pkg)
            pkg_versions.append((pkg, current, latest))

        print(Color.BLUE('Checking requirements latest version done.'))
        print()
        print_table(pkg_versions)

    def generate_reqs(self, save_path, check_path,
                      ignores, comparison_operator):
        gr = GenerateReqs(save_path, check_path, ignores,
                          self.installed_pkgs, comparison_operator)
        gr.generate_reqs()


class GenerateReqs(object):

    def __init__(self, save_path, project_path, ignores,
                 installed_pkgs, comparison_operator='=='):
        self._save_path = save_path
        self._project_path = project_path
        self._ignores = ignores
        self._installed_pkgs = installed_pkgs
        self._maybe_local_mods = set()
        self._comparison_operator = comparison_operator

    def generate_reqs(self):
        """Generate requirements for `project_path`, save file in
        `save_path`.
        """
        print(Color.BLUE('Starting generate requirements ...'))
        reqs, try_imports, guess = self.extract_reqs()
        in_pypi = set()
        answer = 'n'
        pyver = None

        guess.remove(*try_imports)
        if guess:
            print(Color.RED('The following modules are not found yet:'))
            self._invalid_reqs(guess)
            msg = ('Some of them may not install in local environment.\n'
                   'Try to search PyPI for the missing modules and filter'
                   ' some unnecessary modules? (y/[N]) '
                   ).format(pyver)
            sys.stdout.write(Color.RED(msg))
            sys.stdout.flush()
            answer = sys.stdin.readline()
            answer = answer.strip().lower()
            if answer in ('y', 'yes'):
                print(Color.BLUE('Checking modules on the PyPI...'))
                for name, detail in guess.sorted_items():
                    logger.info('Checking %s on the PyPI ...', name)
                    with database() as db:
                        rows = db.query_all(name)
                        pkgs = [row.package for row in rows]
                        if pkgs:
                            in_pypi.add(name)
                            if name in self._maybe_local_mods:
                                self._maybe_local_mods.remove(name)
                        for pkg in self._best_matchs(name, pkgs):
                            latest = check_latest_version(pkg)
                            reqs.add(pkg, latest, detail.comments)

        # Save old requirements file.
        self._save_old_reqs()
        # Write requirements to file.
        self._write_reqs(reqs)
        # If requirements has been covered, show difference.
        self._reqs_diff()
        print(Color.BLUE('Generate requirements done!'))
        del reqs

        if guess and answer in ('y', 'yes'):
            guess.remove(*(in_pypi | self._maybe_local_mods))
            if guess:
                print(Color.RED('These modules are not found:'))
                self._invalid_reqs(guess)
                print(Color.RED('Maybe or you need update database.'))

    def extract_reqs(self):
        """Extract requirements from project."""
        reqs = ReqsModules()
        guess = ReqsModules()
        modules, try_imports, local_mods = project_import_modules(
            self._project_path, self._ignores)
        app_name = os.path.basename(self._project_path)
        if app_name in local_mods:
            local_mods.remove(app_name)

        # Filtering modules
        candidates = self._filter_modules(modules, local_mods)

        logger.info('Check module in local environment.')
        for name in candidates:
            logger.info('Checking module: %s', name)
            if name in self._installed_pkgs:
                pkg_name, version = self._installed_pkgs[name]
                reqs.add(pkg_name, version, modules[name])
            else:
                guess.add(name, 0, modules[name])
        logger.info('Finish local environment checking.')
        return reqs, try_imports, guess

    def _write_reqs(self, reqs):
        print(Color.BLUE('Writing requirements to "{0}"'.format(
            self._save_path)))
        with open(self._save_path, 'w+') as f:
            f.write('# Requirements automatically generated by pigar.\n'
                    '# https://github.com/damnever/pigar\n')
            for k, v in reqs.sorted_items():
                f.write('\n')
                f.write(''.join(['# {0}\n'.format(c)
                                 for c in v.comments.sorted_items()]))
                if k == '-e':
                    f.write('{0} {1}\n'.format(k, v.version))
                elif v:
                    f.write('{0} {1} {2}\n'.format(
                        k, self._comparison_operator, v.version))
                else:
                    f.write('{0}\n'.format(k))

    def _best_matchs(self, name, pkgs):
        # If imported name equals to package name.
        if name in pkgs:
            return [pkgs[pkgs.index(name)]]
        # If not, return all possible packages.
        return pkgs

    def _filter_modules(self, modules, local_mods):
        candidates = set()

        logger.info('Filtering modules ...')
        for module in modules:
            logger.info('Checking module: %s', module)
            if not module or module.startswith('.'):
                continue
            if module in local_mods:
                self._maybe_local_mods.add(module)
            if is_stdlib(module):
                continue
            candidates.add(module)

        return candidates

    def _invalid_reqs(self, reqs):
        for name, detail in reqs.sorted_items():
            print(
                '  {0} referenced from:\n    {1}'.format(
                    Color.YELLOW(name),
                    '\n    '.join(detail.comments.sorted_items())
                )
            )

    def _save_old_reqs(self):
        if os.path.isfile(self._save_path):
            with codecs.open(self._save_path, 'rb', 'utf-8') as f:
                self._old_reqs = f.readlines()

    def _reqs_diff(self):
        if not hasattr(self, '_old_reqs'):
            return
        with codecs.open(self._save_path, 'rb', 'utf-8') as f:
            new_reqs = f.readlines()
        is_diff, diffs = lines_diff(self._old_reqs, new_reqs)
        msg = 'Requirements file has been covered, '
        if is_diff:
            msg += 'there is the difference:'
            print('{0}\n{1}'.format(Color.YELLOW(msg), ''.join(diffs)), end='')
        else:
            msg += 'no difference.'
            print(Color.YELLOW(msg))


def main():
    Main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import re
import codecs

from setuptools import setup, find_packages


version = ''
with open('pigar/_version.py', 'r') as f:
    version = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        f.read(), re.M).group(1)

if not version:
    raise RuntimeError('Cannot find version information')


with codecs.open('README-PYPI.rst', encoding='utf-8') as f:
    long_description = f.read()

with codecs.open('CHANGELOGS.rst', encoding='utf-8') as f:
    change_logs = f.read()

install_requires = [
    'colorama>=0.3.9',
    'requests>=2.20.0',
]
if sys.version_info < (3, 2):
    install_requires.append('futures')


setup(
    name='pigar',
    version=version,
    description=('A fantastic tool to generate requirements for your'
                 ' Python project, and more than that.'),
    long_description=long_description + '\n\n' + change_logs,
    url='https://github.com/damnever/pigar',
    author='damnever',
    author_email='dxc.wolf@gmail.com',
    license='The BSD 3-Clause License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='requirements tool',
    packages=find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pigar=pigar.__main__:main',
        ]
    },
)
